import inspect
from typing import Callable, TypeVar, List

_class_source_info_cache = {}
T = TypeVar('T')


def get_class_source_info(cls):
    if cls not in _class_source_info_cache:
        source_lines, starting_line_no = inspect.getsourcelines(cls)
        file_name = inspect.getfile(cls)
        _class_source_info_cache[cls] = (starting_line_no, starting_line_no + len(source_lines) - 1, file_name)
    return _class_source_info_cache[cls]


def get_calling_context(cls=None):
    caller_frame = inspect.currentframe().f_back.f_back
    file_name = caller_frame.f_code.co_filename
    line_no = caller_frame.f_lineno
    if cls is None:
        local_vars = caller_frame.f_locals
        instance = local_vars['self']
        cls = type(instance)
    start_line_no, end_line_no, class_file_name = get_class_source_info(cls)
    if file_name != class_file_name or not (start_line_no <= line_no <= end_line_no):
        return True
    return False


def private(func):
    def wrapper(*args, **kwargs):
        if get_calling_context():
            raise AttributeError(f"The function '{func.__name__}' is private")
        return func(*args, **kwargs)

    return wrapper


class PrivateMeta(type):
    _private_list: List[str] = []
    _constant_list: List[str] = []

    def __new__(cls, name, bases, dct):

        original_setattr = dct.get('__setattr__', None)
        original_getattr = dct.get('__getattribute__', None)

        for key, value in dct.items():
            if key.startswith('private_'):
                cls._private_list.append(key[8:])
            elif key.startswith('const_'):
                cls._constant_list.append(key[6:])

        def __setattr__(self, item, value):

            if item.startswith('private_'):
                item = item[8:]
                cls._private_list.append(item)

            elif item.startswith('const_'):
                item = item[6:]
                cls._constant_list.append(item)

            position = get_calling_context(self.__class__.__bases__[0])

            if position:
                if item in cls._private_list:
                    raise AttributeError(f"The attribute '{item}' is private")
                elif item in cls._constant_list:
                    raise AttributeError(f"The attribute '{item}' is const")

            if original_setattr:
                original_setattr(self, item, value)
            else:
                super(self.__class__, self).__setattr__(item, value)

        def __getattribute__(self, item):

            if item.startswith('private_'):
                item = item[8:]
            elif item.startswith('const_'):
                item = item[6:]

            if item in cls._private_list and get_calling_context(self.__class__.__bases__[0]):
                raise AttributeError(f"The attribute '{item}' is private")

            if original_getattr:
                return original_getattr(self, item)
            else:
                return object.__getattribute__(self, item)

        def get_private(self):
            return cls._private_list

        def get_const(self):
            return cls._constant_list

        dct['__setattr__'] = __setattr__
        dct['__getattribute__'] = __getattribute__
        dct['get_private'] = get_private
        dct['const_private'] = get_const

        return super().__new__(cls, name, bases, dct)


class PrivateClass:

    def __init__(self, cls: T):
        self.cls = cls

    def __call__(self, *args, **kwargs) -> T:
        cls_dict = dict(self.cls.__dict__)
        cls_dict.pop('__dict__', None)
        cls_dict.pop('__weakref__', None)
        cls_dict.pop('__module__', None)
        cls_dict.pop('__doc__', None)

        Wrapped = PrivateMeta(self.cls.__name__, (self.cls,), cls_dict)
        instance = Wrapped(*args, **kwargs)
        return instance


@PrivateClass
class MyClass:
    private_horse: int
    const_pi: float
    say: Callable

    def __init__(self):
        self.private_horse = 10
        self.const_pi = 3.14

    def private_say(self):
        print("Hello, world!")

    def say_normal(self):
        self.horse = 12
        print(self.horse)
        print("Hello, world!")



import sys
from ctypes import wintypes
import ctypes
import pyautogui

KLF_ACTIVATE = 0x00000001


# 获取当前输入法的语言标识符
def set_english_keyboard_layout():
    hWnd = ctypes.windll.user32.GetForegroundWindow()
    thread_id = ctypes.windll.user32.GetWindowThreadProcessId(hWnd, None)
    layout_id = ctypes.windll.user32.GetKeyboardLayout(thread_id)
    locale_id = 0x0409
    if layout_id & 0xFFFF != locale_id:
        locale_id_wstr = "00000409"
        klid = ctypes.create_unicode_buffer(locale_id_wstr)
        user32 = ctypes.WinDLL("user32")
        LoadKeyboardLayoutW = user32.LoadKeyboardLayoutW
        LoadKeyboardLayoutW.argtypes = [wintypes.LPCWSTR, wintypes.UINT]
        LoadKeyboardLayoutW.restype = wintypes.HKL

        klh = LoadKeyboardLayoutW(klid, KLF_ACTIVATE)
        if klh == 0:
            raise ctypes.WinError()
        print(f"Loaded keyboard layout: {locale_id_wstr}, handle: 0x{klh:016X}")
        ctypes.windll.user32.ActivateKeyboardLayout(klh, KLF_ACTIVATE)
        hwnd = ctypes.windll.user32.GetForegroundWindow()
        thread_id = ctypes.windll.user32.GetWindowThreadProcessId(hwnd, None)
        hkl = ctypes.windll.user32.LoadKeyboardLayoutW(f"{locale_id:08X}", KLF_ACTIVATE)
        if hkl == 0:
            raise ctypes.WinError("Failed to load keyboard layout")
        ctypes.windll.user32.PostMessageW(hwnd, 0x0050, 0, hkl)
    else:
        # print("already english")
        pass
    return


def ensure_english_keyboard():
    set_english_keyboard_layout()


def is_caps_lock_on():
    hllDll = ctypes.WinDLL("User32.dll")
    VK_CAPITAL = 0x14
    return hllDll.GetKeyState(VK_CAPITAL) & 0x0001


def switch_to_lowercase():
    pyautogui.press('capslock')
    # print("Switched to lowercase by sending Shift key.")


def ensure_lowercase():
    if is_caps_lock_on():
        # print("Caps Lock is on, switching to lowercase...")
        switch_to_lowercase()
    else:
        pass
        # print("Caps Lock is already off, no need to switch.")


def block_input(block):
    ctypes.windll.user32.BlockInput(block)


def ensure_enter(func):
    def wrapper(*args, **kwargs):
        ensure_lowercase()
        ensure_english_keyboard()
        block_input(True)
        result = func(*args, **kwargs)
        block_input(False)
        # print(f"Function '{func.__name__}' execute with ensure enter")
        return result

    return wrapper


if __name__ == "__main__":
    print("Python {0:s} {1:d}bit on {2:s}\n".format(" ".join(item.strip() for item in sys.version.split("\n")),
                                                    64 if sys.maxsize > 0x100000000 else 32, sys.platform))
    ensure_lowercase()



import functools
import time
from typing import Optional, Callable, Union
import line_profiler


def timeit(func, min_time=0):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        if elapsed_time >= min_time:
            print(f"Function '{func.__name__}' took {end_time - start_time} seconds to execute.")
        return result

    return wrapper


class TimeIt:
    start_time: float
    end_time: float
    execution_time: float
    round: Optional[int]

    def __init__(self, round: Optional[int] = None, by_line: bool = False, output_unit='auto'):
        self.round = round
        self.logger = []
        self.by_line = by_line
        self.profiler = line_profiler.LineProfiler() if by_line else None
        self.output_unit = output_unit

    def _init(self, round: Optional[int] = None, by_line: bool = None, output_unit=None):
        if round is not None:
            self.round = round
        if by_line is not None:
            self.by_line = by_line
        if output_unit is not None:
            self.output_unit = output_unit

    def add(self, func):
        if self.by_line and self.profiler:
            self.profiler.add_function(func)

    def __enter__(self):
        self.start_time = time.time()
        if self.by_line:
            self.profiler.enable_by_count()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.by_line and self.profiler:
            self.profiler.disable_by_count()
        self.execution_time = time.time() - self.start_time
        if self.output_unit == 'auto':
            if self.execution_time >= 1:
                unit = 1
            elif self.execution_time >= 1e-3:
                unit = 1e-3
            else:
                unit = 1e-6
        else:
            unit = self.convert_output_unit(self.output_unit)
        if self.round is not None:
            formatted_time = round(self.execution_time, self.round)
        else:
            formatted_time = self.execution_time
        print(
            f"Execution time: {formatted_time:.{self.round}f} seconds" if self.round is not None else f"Execution time: {formatted_time} seconds")
        self.logger.append(self.execution_time)
        if self.by_line and self.profiler:
            self.profiler.print_stats(output_unit=unit)
            self.profiler = line_profiler.LineProfiler()

    @staticmethod
    def convert_output_unit(unit: Union[str, float]) -> float:
        if unit == 's':
            return 1
        elif unit == 'ms':
            return 1e-3
        elif unit == 'us':
            return 1e-6
        else:
            return unit

    def clear(self):
        self.logger.clear()
        self.profiler = line_profiler.LineProfiler()

    def clear_profiler(self):
        self.profiler = line_profiler.LineProfiler()

    def list_comprehension(self):
        numbers = self.logger
        return [current - previous for previous, current in zip(numbers[:-1], numbers[1:])]

    def profile_function(self, enable_timing: bool = False) -> Callable:
        """Decorator to profile a function with line_profiler if by_line is True."""

        def decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                if enable_timing:
                    with self:
                        result = func(*args, **kwargs)
                else:
                    result = func(*args, **kwargs)
                return result

            if self.by_line and self.profiler:
                self.profiler.add_function(func)

            return wrapper

        return decorator

    def __call__(self, round: Optional[int] = None, by_line: bool = None, output_unit=None):
        self._init(round, by_line, output_unit)
        return self



import subprocess
import webbrowser
import time
import pygetwindow as gw
from screeninfo import get_monitors
import pyautogui
import pyperclip
from .check_input import ensure_enter


class MyChrome:

    def __init__(self, name="", window=None):
        self.name = name
        self.window = window
        self.screens_info = get_screens_info()
        self.window_info = get_window_info(self.window)
        self.current_screen = self.get_window_position()

    def open(self, url, new=0):
        print(url, new, self.window)
        if new == 0 and self.window is not None:
            self.window.activate()
            webbrowser.open(url)
        elif new == 1 and self.window is not None:
            self.window.activate()
            webbrowser.open(url)
        elif new == 2 or self.window is None:
            before_windows = gw.getAllWindows()

            subprocess.Popen([r"C:\Program Files\Google\Chrome\Application\chrome.exe", '--new-window', 'google.com'])

            new_window = None
            for _ in range(30):
                after_windows = gw.getAllWindows()
                new_window = next((window for window in after_windows if window not in before_windows), None)
                if new_window:
                    break
                time.sleep(0.5)
            new_window.maximize()
            self.window = new_window
            self.window_info = get_window_info(self.window)
            self.current_screen = self.get_window_position()
            self.window.activate()
            webbrowser.open(url)

    def close(self):
        self.window.activate()
        pyautogui.hotkey('ctrl', 'w')

    def set_window(self, window):
        self.window = window

    @ensure_enter
    def put(self, cmd):
        pyautogui.hotkey('ctrl', 'shift', 'j')
        time.sleep(.2)
        # pyautogui.press('tab')
        pyperclip.copy(cmd)
        pyautogui.hotkey('ctrl', 'v')
        pyautogui.press('enter')
        pyautogui.hotkey('F12')

    @staticmethod
    def write(script):
        pyperclip.copy(script)
        pyautogui.hotkey('ctrl', 'v')
        time.sleep(0.1)
        pyautogui.press('enter')

    def get_window_position(self):
        return which_screen_is_window_on(self.window_info, self.screens_info)

    def move_window_to_screen(self, screen_choice, maximize=True, center=False):
        if self.current_screen == screen_choice or self.current_screen == -1:
            return None
        move_window_to_screen(self.window, self.screens_info[screen_choice], maximize=maximize, center=center)


def get_screens_info():
    screens = []
    for monitor in get_monitors():
        screens.append({
            'left': monitor.x,
            'top': monitor.y,
            'width': monitor.width,
            'height': monitor.height,
            'right': monitor.x + monitor.width,
            'bottom': monitor.y + monitor.height
        })
    return screens


def get_window_info(window):
    if window:
        return {
            'left': window.left,
            'top': window.top,
            'width': window.width,
            'height': window.height,
            'right': window.left + window.width,
            'bottom': window.top + window.height
        }
    return None


def calculate_overlap_area(window_info, screen_info):
    overlap_left = max(window_info['left'], screen_info['left'])
    overlap_top = max(window_info['top'], screen_info['top'])
    overlap_right = min(window_info['right'], screen_info['right'])
    overlap_bottom = min(window_info['bottom'], screen_info['bottom'])

    if overlap_left < overlap_right and overlap_top < overlap_bottom:
        return (overlap_right - overlap_left) * (overlap_bottom - overlap_top)
    return 0


def which_screen_is_window_on(window_info, screens_info):
    if window_info is None:
        return -1
    max_overlap_area = 0
    best_screen_idx = -1

    for idx, screen in enumerate(screens_info):
        overlap_area = calculate_overlap_area(window_info, screen)
        if overlap_area > max_overlap_area:
            max_overlap_area = overlap_area
            best_screen_idx = idx

    return best_screen_idx


def move_window_to_screen(window, screen_info, maximize=True, center=False):
    window.restore()
    if window:
        if center:
            new_left = screen_info['left'] + (screen_info['width'] - window.width) // 2
            new_top = screen_info['top'] + (screen_info['height'] - window.height) // 2
            print(f"窗口已移动到屏幕的中间: ({new_left}, {new_top})")
        else:
            new_left = screen_info['left']
            new_top = screen_info['top']
            print(f"窗口已移动到屏幕的左上角: ({new_left}, {new_top})")
        # if maximize and not window.isMaximized:
        #     window.maximize()
        window.moveTo(new_left, new_top)
        if maximize and not window.isMaximized:
            window.maximize()
        print(window.isMaximized)
        print(window.width, window.height)



import subprocess
import sys
import os


def update_environment(requirements_file='requirements.txt'):
    try:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', '-r', requirements_file])
        print("Environment updated successfully.")
    except subprocess.CalledProcessError as e:
        print(f"An error occurred while updating the environment: {e}")


def main():
    from .file_config import find_top_level_package
    package = find_top_level_package(os.getcwd())
    requirements_file = os.path.join(package, 'requirements.txt')
    if not os.path.exists(requirements_file):
        print('No requirements file found. Please create one and try again.')
        return None
    print(requirements_file)
    update_environment(requirements_file)


if __name__ == "__main__":
    update_environment()



import argparse
import os
import queue
import signal
import subprocess
import sys
import threading


def enqueue_output(out, q):
    try:
        for line in iter(out.readline, ''):
            q.put(line)
        out.close()
    except Exception as e:
        print(f"Error reading output: {str(e)}")


def run_command(directory, command, pid_queue=None):
    try:
        # 切换到指定目录
        os.chdir(directory)
        print(f"Running '{command}' in {directory}")
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True,
                                   encoding='utf-8', errors='replace')
        print(process.pid)
        if pid_queue is not None:

            pid_queue.put(process.pid)

        q_stdout = queue.Queue()
        q_stderr = queue.Queue()

        t_stdout = threading.Thread(target=enqueue_output, args=(process.stdout, q_stdout))
        t_stderr = threading.Thread(target=enqueue_output, args=(process.stderr, q_stderr))

        t_stdout.start()
        t_stderr.start()
        print('start')

        while t_stdout.is_alive() or t_stderr.is_alive() or not q_stdout.empty() or not q_stderr.empty():
            while not q_stdout.empty():
                line = q_stdout.get_nowait()
                print(f"{directory}: {line}", end="")
            while not q_stderr.empty():
                line = q_stderr.get_nowait()
                print(f"{directory} [ERROR]: {line}", end="")
        t_stdout.join()
        t_stderr.join()

        return_code = process.wait()
        if return_code:
            raise subprocess.CalledProcessError(return_code, command)
    except subprocess.CalledProcessError as e:
        print(f"An error occurred while running '{command}' in {directory}: {e}")
    except Exception as e:
        print(f"An unexpected error occurred while running '{command}' in {directory}: {str(e)}")




import inspect


class ClassBuilder:

    @staticmethod
    def get_params():
        frame = inspect.currentframe().f_back
        args, _, _, values = inspect.getargvalues(frame)
        params = {arg: values[arg] for arg in args[1:]}
        return params

    @staticmethod
    def auto_init(obj):
        frame = inspect.currentframe().f_back
        args, _, _, values = inspect.getargvalues(frame)
        for arg in args[1:]:
            setattr(obj, arg, values[arg])





# class A:
#
#     def __init__(self, name, home):
#         ClassBuilder.auto_init(self)
#         print(self.name)



import os
import sys

import inspect
import os
import pkgutil


def find_top_level_package(file_path=None):
    if file_path is None:
        file_path = os.getcwd()
    current_dir = file_path
    below_dir = current_dir
    while current_dir:
        if '__init__.py' in os.listdir(current_dir):
            parent_dir = os.path.dirname(current_dir)
            if parent_dir == current_dir:  # 根目录
                break
            below_dir = current_dir
            current_dir = parent_dir
        else:
            break
    return below_dir


def main():
    print(find_top_level_package(os.getcwd()))


if __name__ == '__main__':
    main()



from typing import List, Union, TypeVar, Callable

from typing import Callable


def restruct_decorator(func_with_old_decorator: Callable, old_decorator: Callable, new_decorator: Callable) -> Callable:
    """
    Restructures a decorator by applying a new decorator to an existing function while maintaining the original function's behavior.

    Args:
    func_with_old_decorator (Callable): The function to which the old decorator is applied.
    old_decorator (Callable): The old decorator to be removed.
    new_decorator (Callable): The new decorator to be applied.

    Returns:
    Callable: The function with the old decorator removed and the new decorator applied.
    """
    original_func = func_with_old_decorator.__closure__[0].cell_contents
    return old_decorator(new_decorator(original_func))


def add_decorator(decorator: Callable, func:Callable) -> Callable:
    return decorator(func)




import numpy as np
from abc import ABC, abstractmethod
import random


class Game(ABC):
    choice = random.choice

    @abstractmethod
    def __init__(self):
        pass

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def get_legal_actions(self):
        pass

    @abstractmethod
    def step(self, action):
        pass

    @abstractmethod
    def check_winner(self):
        pass

    def simulate(self):
        states, rewards, next_states, actions, dones = [], [], [], [], []

        while self.check_winner() == 0:
            states.append(self.state())
            actions = self.get_legal_actions()
            action = self.choice(actions)
            board, reward, done = self.step(action)
            rewards.append(reward)
            next_states.append(board)
            actions.append(action)
            dones.append(done)

        return states, actions, rewards, next_states, dones

    @abstractmethod
    def state(self):
        pass

    @abstractmethod
    def transfer_action(self):
        pass

# class Gomoku(Game):
#
#     def __init__(self, board_size=19, win_length=5):
#         super().__init__()
#         return
#
#
# if __name__ == '__main__':
#     game = Gomoku()



from .setup_paths import setup_paths, log_call_position, find_print, find_function
from .custom_importer import import_from, import_module, get_location, find_spec
from .AI import AI
from .Discord import Discord
from .Argument import Argument
from .class_builder import ClassBuilder



import os
from tqdm import tqdm


def find_file(directory: str, filename: str, ignore_extension: bool = True, verbose: bool = True):
    if ignore_extension and '.' in filename:
        filename, _ = os.path.splitext(filename)
    find_result = []
    for root, dirs, files in tqdm(os.walk(directory)):
        if ignore_extension:
            for file in files:
                file_name, file_ext = os.path.splitext(file)
                if file_name == filename:
                    find_result.append(os.path.join(root, file))
                    if verbose:
                        print(os.path.join(root, file))
        else:
            if filename in files:
                find_result.append(os.path.join(root, filename))
                if verbose:
                    print(os.path.join(root, filename))

    return find_result



import os
import random
from collections import deque
import numpy as np
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from .AI import Models, Optimizer
import torch
import copy


class DQN:

    def __init__(self, game=None, q_network=None, target_network=None, optimizer=None, replay=None, criterion=None,
                 lr=1e-5, gamma=0.995, target_update=50, batch_size=64, epsilon_start=1.0, epsilon_end=0.01,
                 epsilon_decay=0.995, capacity = 10000,
                 log_dir=None, writer=None, device=None, num_classes=9, epoch=0
                 ):
        if replay is None:
            replay = MemoryReplay(capacity=capacity)
        if q_network is None:
            q_network = Models.resnet18()
            num_ftrs = q_network.fc.in_features
            q_network.fc = torch.nn.Linear(num_ftrs, num_classes)
        if target_network is None:
            target_network = Models.resnet18()
            num_ftrs = target_network.fc.in_features
            target_network.fc = torch.nn.Linear(num_ftrs, num_classes)
        if optimizer is None:
            optimizer = Optimizer.Adam(q_network.parameters(), lr=lr)
        if criterion is None:
            criterion = torch.nn.MSELoss()
        if device is None:
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.lr = lr
        self.gamma = gamma
        self.game = game
        self.epsilon_start = epsilon_start
        self.epsilon_end = epsilon_end
        self.epsilon_decay = epsilon_decay
        self.target_update = target_update
        self.batch_size = batch_size
        self.replay = replay
        self.q_network = q_network.to(device)
        self.target_network = target_network.to(device)
        self.target_network.eval()
        self.optimizer = optimizer
        self.criterion = criterion
        self.device = device
        self.epoch = epoch
        if writer is None:
            writer = SummaryWriter(log_dir)
        self.writer = writer
        self.log_dir = log_dir if log_dir is not None else writer.get_logdir()
        self.epsilon = epsilon_start

    def prefill_replay_buffer(self, prefill_steps):
        state = self.game.reset()
        for _ in range(prefill_steps):
            action = random.choice(self.game.get_legal_actions())
            next_state, reward, done, _ = self.game.step(action)
            next_state = np.array(next_state)
            action = self.game.transfer_action(action)
            self.replay.push(state, action, reward, next_state, done)
            state = next_state if not done else self.game.reset()

    def train(self, episodes):
        device = self.device
        epsilon = self.epsilon
        self.q_network.train()
        for episode in tqdm(range(episodes), desc="Training Episodes"):
            start_player = 1 if random.random() < 0.5 else 2
            state = self.game.reset(start_player=start_player)
            state = np.array(state)
            total_reward = 0

            for t in range(200):
                if random.random() < epsilon:
                    action = random.choice(self.game.get_legal_actions())
                else:
                    with torch.no_grad():
                        board = torch.FloatTensor(state)
                        board = board.unsqueeze(0).to(device)
                        self.q_network.eval()
                        action_values = self.q_network(board).squeeze(0)
                        legal_actions_pos = self.game.get_legal_actions()
                        legal_actions = [self.game.transfer_action(action) for action in legal_actions_pos]
                        arg = torch.argmax if self.game.current_player == start_player else torch.argmin
                        action = legal_actions_pos[arg(action_values[legal_actions]).item()]
                        self.q_network.train()

                next_state, reward, done, _ = self.game.step(action)
                next_state = np.array(next_state)
                action = self.game.transfer_action(action)
                self.replay.push(state, action, reward, next_state, done)

                state = next_state
                total_reward += reward

                if len(self.replay) >= self.batch_size:
                    states, actions, rewards, next_states, dones = self.replay.sample(self.batch_size)

                    states = torch.FloatTensor(states).reshape(self.batch_size, 3, 3, 3).to(device)
                    actions = torch.LongTensor(actions).unsqueeze(1).to(device)
                    rewards = torch.FloatTensor(rewards).to(device)
                    next_states = torch.FloatTensor(next_states).reshape(self.batch_size, 3, 3, 3).to(device)
                    dones = torch.FloatTensor(dones).to(device)

                    q_values = self.q_network(states).gather(1, actions).squeeze()
                    next_q_values = self.target_network(next_states).max(1)[0]
                    target_q_values = rewards + self.gamma * next_q_values * (1 - dones)

                    loss = self.criterion(q_values, target_q_values.detach())
                    self.optimizer.zero_grad()
                    loss.backward()

                    # 梯度裁剪
                    torch.nn.utils.clip_grad_norm_(self.q_network.parameters(), max_norm=0.5)
                    self.optimizer.step()

                    self.writer.add_scalar('Reward', total_reward, self.epoch + episode)
                    self.writer.add_scalar('Epsilon', epsilon, self.epoch + episode)
                    self.writer.add_scalar('Loss', loss.item(), self.epoch + episode)

                if done:
                    break

            epsilon = max(self.epsilon_end, self.epsilon_decay * epsilon)

            if episode % self.target_update == 0:
                self.target_network.load_state_dict(self.q_network.state_dict())

                print(f"Episode {episode}, Total Reward: {total_reward}, Epsilon: {epsilon:.2f}")

        self.epoch += episodes
        self.writer.close()

    def test_model(self, episodes=10):
        device = self.device
        self.q_network.eval()
        total_rewards = []

        for episode in range(episodes):
            state = self.game.reset()
            state = np.array(state)
            total_reward = 0

            while True:
                with torch.no_grad():
                    board = torch.FloatTensor(state)
                    board = board.unsqueeze(0).to(device)
                    action_values = self.q_network(board).squeeze(0)
                    legal_actions_pos = self.game.get_legal_actions()
                    legal_actions = [self.game.transfer_action(action) for action in legal_actions_pos]
                    arg = torch.argmax if self.game.current_player == 1 else torch.argmin

                    action = legal_actions_pos[arg(action_values[legal_actions]).item()]

                next_state, reward, done, winner = self.game.step(action)
                next_state = np.array(next_state)
                state = next_state
                total_reward += reward

                if done:
                    break

            total_rewards.append(total_reward)
            print(f"Test Episode {episode + 1}, Total Reward: {total_reward}, Winner: {winner}")

        avg_reward = np.mean(total_rewards)
        print(f"Average Total Reward over {episodes} episodes: {avg_reward}")

    def save(self, path):
        if '.pth' in path:
            return self._save_config(path)
        os.makedirs(path, exist_ok=True)
        return self._save_config(os.path.join(path, 'dqn.pth'))

    def _save_config(self, path):
        try:
            torch.save({'game': self.game, 'q_network': self.q_network, 'target_network': self.target_network,
                        'epoch': self.epoch, 'optimizer': self.optimizer, 'criterion': self.criterion, 'lr': self.lr,
                        'gamma': self.gamma, 'device': self.device, 'log_dir': self.log_dir, 'replay': self.replay,
                        'target_update': self.target_update, 'batch_size': self.batch_size,
                        'epsilon_start': self.epsilon,
                        'epsilon_end': self.epsilon_end, 'epsilon_decay': self.epsilon_decay
                        },
                       path)
            print('Saved model configuration successfully')
            return True
        except Exception as e:
            print(f'Error saving model configuration: {str(e)}')
            return False

    @staticmethod
    def load(path):
        if not os.path.exists(path):
            raise FileNotFoundError(f"No checkpoint found at '{path}'")
        params = torch.load(os.path.join(path, 'dqn.pth'))
        return DQN(**params)


class SAC:

    def __init__(self, model, game, replay=None):
        self.model = model
        self.game = game
        if replay is None:
            replay = MemoryReplay()
        self.replay = replay

        pass

    def get_data(self):
        for i in range(200):
            res = self.game.simulate()
            self.replay.push_list(*res)
            self.game.reset()


class MemoryReplay:
    def __init__(self, capacity=10000):
        self.capacity = capacity
        self.memory = deque(maxlen=capacity)

    def push(self, state, action, reward, next_state, done):
        experience = (state, action, reward, next_state, done)
        self.memory.append(experience)

    def push_list(self, states, actions, rewards, next_states, dones):
        for state, action, reward, next_state, done in zip(states, actions, rewards, next_states, dones):
            self.push(state, action, reward, next_state, done)

    def sample(self, batch_size):
        experiences = random.sample(self.memory, batch_size)

        states, actions, rewards, next_states, dones = zip(*experiences)

        return (np.array(states),
                np.array(actions),
                np.array(rewards, dtype=np.float32),
                np.array(next_states),
                np.array(dones, dtype=np.uint8))

    def __len__(self):
        return len(self.memory)

    def size(self):
        return self.__len__()



import importlib
import importlib.util


def import_from(module, function):
    md = importlib.import_module(module)
    cls = getattr(md, function)
    return cls


def import_module(name, package=None):
    return importlib.import_module(name, package)


def get_location(module):
    spec = importlib.util.find_spec(module)
    return spec.origin


find_spec = importlib.util.find_spec



import math
import random
import sys
from typing import List, Union, TypeVar, Callable

T = TypeVar('T')


def truncate_with_floor(number: float, digits: int = 0):
    if digits == 0:
        return math.floor(number)
    stepper = 10 ** digits
    return math.floor(number * stepper) / stepper


def product(arr: List[Union[int, float]]) -> Union[int, float]:
    if sys.version_info >= (3.8, 0):
        return math.prod(arr)
    else:
        from functools import reduce
        return reduce(lambda x, y: x * y, arr)


def choose(best_choice: Callable[[List[T]], T], actions: List[T], epsilon: float = 0.5) -> T:
    return random.choice(actions) if random.random() < epsilon else best_choice(actions)


def random_choice(array: List[T]) -> T:
    return random.choice(array)


def probability(value: float) -> bool:
    rand_number = random.uniform(0, 1)
    return rand_number <= value


def set_minus(set1: List[T], set2: List[T]) -> List[T]:
    return [i for i in set1 if i not in set2]


if __name__ == '__main__':
    from performer_helper import TimeIt

    with TimeIt():
        print(product([1, 2, 3, 4, 5, 6, 7]))



import os
import queue
import threading
from .utils import run_command


def run_cmd(command, *args, **kwargs):
    current_directory = os.getcwd()
    print("Current Directory:", current_directory)
    StarRail = current_directory

    pid_queue = queue.Queue()
    cmd = command + " " + " ".join(args)
    print(cmd)
    commands = [
        (StarRail, cmd)
    ]
    try:
        print("Running... Press Ctrl+C to interrupt")
        threads = []

        for i, (directory, command) in enumerate(commands):
            thread = threading.Thread(target=run_command, args=(directory, command, pid_queue))
            threads.append(thread)
            thread.start()
            thread.join()

            print("Command completed. Clearing environment before starting the next command...")
            # clear_environment()

    except SystemExit:
        print("Process terminated gracefully")


def run(commands):
    pid_queue = queue.Queue()
    try:
        print("Running... Press Ctrl+C to interrupt")
        threads = []

        for i, (directory, command) in enumerate(commands):
            thread = threading.Thread(target=run_command, args=(directory, command, pid_queue))
            threads.append(thread)
            thread.start()
            thread.join()

            print("Command completed. Clearing environment before starting the next command...")
            # clear_environment()

    except SystemExit:
        print("Process terminated gracefully")


def get_branch(*args, **kwargs):
    run([(os.getcwd(), 'git branch')])



import logging
from typing import Optional
import discord
from discord.utils import MISSING


class DiscordClient(discord.Client):

    def __init__(self, *, intents=None, token=None, **options):
        if intents is None:
            intents = discord.Intents.all()
        self.token = token
        self.message_logger = []
        super().__init__(intents=intents, **options)

    async def on_ready(self):
        print('Connected to bot: {}'.format(self.user.name))
        print('Bot ID: {}'.format(self.user.id))

    async def on_message(self, message):
        self.message_logger.append(message)

        if message.author == self.user:
            return

        await self.message(message)

    async def message(self, message):
        pass

    def run(
            self,
            token: str = None,
            *,
            reconnect: bool = True,
            log_handler: Optional[logging.Handler] = MISSING,
            log_formatter: logging.Formatter = MISSING,
            log_level: int = MISSING,
            root_logger: bool = False,
    ) -> None:
        if token is None:
            token = self.token
        try:
            super().run(token, reconnect=reconnect, log_handler=log_handler,
                        log_formatter=log_formatter, log_level=log_level, root_logger=root_logger)
        except SystemExit:
            print("Process terminated gracefully")



import inspect
import sys

import pyperclip
import pyautogui
import datetime
import time as _time
import os
from .MyBrowser import MyChrome


class AutoControl:

    @staticmethod
    def copy(source):
        return pyperclip.copy(source)

    @staticmethod
    def paste():
        return pyautogui.hotkey('ctrl', 'v')

    @staticmethod
    def get_time():
        return datetime.datetime.now()

    @staticmethod
    def shutdown():
        time = sys.argv[1] if len(sys.argv) > 1 else 0
        return os.system(f'shutdown /s /f /t {time}')

    @staticmethod
    def open(file):
        return os.system(f'start {file}')

    @staticmethod
    def open_url(url):
        browser = MyChrome()
        browser.open(url)
        return browser


class Timer:
    def __init__(self, current_time=None):
        self.current_time = datetime.datetime.now() if current_time is None else current_time
        self._interval = 0.1

    @property
    def interval(self):
        return self._interval

    @staticmethod
    def sleep(number):
        return _time.sleep(number)

    def set_task(self, timer, func):
        self.wait_timer(timer)

        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    def wait_timer(self, timer):
        while datetime.datetime.now() < timer:
            _time.sleep(self.interval)
        print(timer)


class Datetime(datetime.datetime):

    def __new__(cls, year=None, month=None, day=None, hour=0, minute=0, second=0,
                microsecond=0, tzinfo=None, *, fold=0):
        now = datetime.datetime.now()
        if year is None:
            year = now.year
        if month is None:
            month = now.month
        if day is None:
            day = now.day

        return super().__new__(cls, year=year, month=month, day=day, hour=hour, minute=minute, second=second,
                               microsecond=microsecond, tzinfo=tzinfo, fold=fold)

    @classmethod
    def tomorrow(cls):
        t = _time.time() + 24 * 60 * 60
        return cls.fromtimestamp(t)

    def next(self, time):
        t = self.timestamp() + time
        return self.fromtimestamp(t)



import sys
from multiprocessing import Process, Lock
import time
import os
import runpy


def child_process():
    print(f"Child process {os.getpid()} started")
    time.sleep(2)
    print(f"Child process {os.getpid()} finished")


def parent_process():
    print(sys.argv)
    print(f"Parent process {os.getpid()} started")
    p = Process(target=child_process)
    p.start()
    print(f"Parent process {os.getpid()} created child process {p.pid}")
    p.join()  # 等待子进程结束
    print(f"Parent process {os.getpid()} finished")


def run(func):
    def wrapper(*args, **kwargs):
        p = Process(target=func, args=args, kwargs=kwargs)
        p.start()
        print("Main process started")
        p.join()

    return wrapper


def main():
    parent_process()


if __name__ == "__main__":
    main()



import re
from typing import List


def get_number(string: str) -> List[float]:
    nums = re.findall(r'\d+\.?\d*', string)
    return [float(x) for x in nums]


def split_string(s: str) -> str:
    words = re.findall(r'[A-Z][a-z]*', s)
    prepositions = {'the', 'of', 'for', 'and', 'in', 'at', 'on', 'to'}
    added = False
    dictionary = ['Coff', 'Cruisg']
    final_words = []
    for word in words:
        for p in prepositions:
            if p in word and (word.startswith(p) or word.endswith(p)):
                if word.startswith(p):
                    remain = word.replace(p, "", 1)
                    final_words.append(p)
                    final_words.append(split_string(remain))
                    added = True
                    break
                elif word.endswith(p):
                    remain = word.replace(p, "", 1)
                    if remain[-1] in ['a', 'i', 'o', 'u'] or remain in dictionary or (
                            remain[-1] in ['e', 'i'] and remain[-2] in ['a', 'i', 'o', 'u', 'A', 'O', 'U', 'I', 'w']):
                        continue
                    final_words.append(split_string(remain))
                    final_words.append(p)
                    added = True
                    break
                else:
                    pass
        if added:
            added = False
            continue
        final_words.append(word)

    return ' '.join(final_words)


def convert_lightcone_effect(text: str, rank: int) -> str:
    pattern = r'(\d+(\.\d+)?(%?)(/\d+(\.\d+)?%?){4})'
    matches = [match[0] for match in re.finditer(pattern, text)]
    if len(matches) > 0:
        for match_obj in matches:
            values = match_obj.split('/')
            text = re.sub(match_obj, values[rank - 1], text)
    return text


def is_numeric(s: str) -> bool:
    if not isinstance(s, str):
        return False
    pattern = r'^[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?$'
    return bool(re.match(pattern, s))


def get_string(s: str) -> str:
    return s



class BidirectionalDict(dict):
    def __init__(self, *args, **kwargs):
        super(BidirectionalDict, self).__init__()
        for key, value in dict(*args, **kwargs).items():
            self[key] = value

    def __setitem__(self, key, value):
        if key in self:
            del self[key]
        if value in self:
            del self[value]
        super(BidirectionalDict, self).__setitem__(key, value)
        super(BidirectionalDict, self).__setitem__(value, key)

    def __getitem__(self, key):
        if key in self:
            return super().__getitem__(key)
        import re
        stripped_key = re.sub(r'\W+', '', key)
        if stripped_key in self:
            return super().__getitem__(stripped_key)
        else:
            return super().__getitem__(key)

    def __delitem__(self, key):
        value = self[key]
        super(BidirectionalDict, self).__delitem__(key)
        super(BidirectionalDict, self).__delitem__(value)


import argparse


class ArgumentParser:

    def __init__(self, description=None, auto=True):
        if description is None:
            description = "Processing some arguments"
        self.parser = argparse.ArgumentParser(description=description)
        self.auto = auto
        self.kwargs = {}
        self.args = []

    def auto_setting(self):
        self.parser.add_argument("args", nargs='*', help="Positional arguments")
        self.parser.add_argument('--kwargs', nargs='*', help='Keyword arguments in the form key=value', default={})

    def add_argument(self, *args, **kwargs):
        self.parser.add_argument(*args, **kwargs)

    def add_arg(self, name, type, help=None):
        self.parser.add_argument(name, type=type, help=help)

    def parse_args(self):
        if self.auto:
            self.auto_setting()
        args = self.parser.parse_args()
        kwargs = {}
        for kwarg in args.kwargs:
            key, value = kwarg.split('=')
            kwargs[key] = value

        for k,v in vars(args).items():
            if k not in ['args', 'kwargs']:
                kwargs[k] = v
        self.args = args.args
        self.kwargs = kwargs
        return None

    def get(self, key):
        return self.kwargs.get(key, None)

    def __repr__(self):
        return f'args: {self.args} kwargs: {self.kwargs}'

    def __getitem__(self, key):
        return self.kwargs.get(key, None)



import os
import subprocess
import argparse
import sys
import queue
import threading
import signal


def get_paramsdict():
    parser = argparse.ArgumentParser(description='Process some text.')
    parser.add_argument("filename", type=str, help="Given a filename")
    parser.add_argument("args", nargs='*', help="Positional arguments")
    parser.add_argument('--kwargs', nargs='*', help='Keyword arguments in the form key=value', default={})

    args = parser.parse_args()

    kwargs = {}
    for kwarg in args.kwargs:
        key, value = kwarg.split('=')
        kwargs[key] = value

    return vars(args)


def run_python(filename, *args, **kwargs):
    if not filename.endswith('.py'):
        filename += '.py'
    abs_filename = os.path.abspath(filename)
    if not os.path.exists(abs_filename):
        print(f"File {abs_filename} does not exist")
        from .file_processing import find_file
        files = find_file(os.getcwd(), filename, False)
        if len(files) == 1:
            abs_filename = os.path.abspath(files[0])
            print(f"Found {abs_filename}")
        else:
            print(f"Found {len(files)} files matching {filename}")
            print(files)
            return

    command = [sys.executable, "-c", f"""
from my_import.setup_paths import setup_paths
setup_paths(verbose=False)
import runpy
runpy.run_path(r'{abs_filename}', run_name='__main__')
"""]

    for arg in args:
        command.append(str(arg))

    for key, value in kwargs.items():
        command.append(f"--{key}")
        command.append(str(value))

    try:
        result = subprocess.run(command, capture_output=True, text=True, check=True)
        if len(result.stdout) == 0:
            return
        print("Output:")
        print(result.stdout)
        if len(result.stderr) == 0:
            return
        print("Errors:")
        print(result.stderr)
    except subprocess.CalledProcessError as e:
        print(f"Error running {filename}: {e}")
        print("Output:")
        print(e.stdout)
        print("Errors:")
        print(e.stderr)


def main(params=None):
    if params is None:
        params = get_paramsdict()
    run_python(params.get('filename'), *params.get('args'), **params.get('kwargs'))


class Runner:

    def __init__(self):
        self.pid_queue = queue.Queue()

    def run(self, commands):
        signal.signal(signal.SIGINT, self.signal_handler)
        try:
            print("Running... Press Ctrl+C to interrupt")
            threads = []

            for i, (directory, command) in enumerate(commands):
                if directory is None:
                    directory = os.getcwd()
                thread = threading.Thread(target=self.run_command, args=(directory, command, self.pid_queue))
                threads.append(thread)
                thread.start()
                thread.join()

                print("Command completed. Clearing environment before starting the next command...")

            while not self.pid_queue.empty():
                id = self.pid_queue.get()
                try:
                    os.kill(id, signal.SIGTERM)
                    print(f"Killing process {id}")
                except OSError:
                    print(f"Process {id} already terminated")

        except SystemExit:
            print("Process terminated gracefully")

    def signal_handler(self, sig, frame):
        print('You pressed Ctrl+C!')
        print(self.pid_queue)
        while not self.pid_queue.empty():
            id = self.pid_queue.get()
            try:
                os.kill(id, signal.SIGTERM)
                print(f"Killing process {id}")
            except OSError:
                print(f"Process {id} already terminated")
        sys.exit(0)

    @staticmethod
    def enqueue_output(out, q):
        try:
            for line in iter(out.readline, ''):
                q.put(line)
            out.close()
        except Exception as e:
            print(f"Error reading output: {str(e)}")

    def run_command(self, directory, command, pid_queue=None):
        try:
            # 切换到指定目录
            os.chdir(directory)
            print(f"Running '{command}' in {directory}")
            process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                       text=True,
                                       encoding='utf-8', errors='replace', cwd=directory)
            print(process.pid)
            if pid_queue is not None:
                pid_queue.put(process.pid)

            q_stdout = queue.Queue()
            q_stderr = queue.Queue()

            t_stdout = threading.Thread(target=Runner.enqueue_output, args=(process.stdout, q_stdout))
            t_stderr = threading.Thread(target=Runner.enqueue_output, args=(process.stderr, q_stderr))

            t_stdout.start()
            t_stderr.start()
            print('start')

            while t_stdout.is_alive() or t_stderr.is_alive() or not q_stdout.empty() or not q_stderr.empty():
                while not q_stdout.empty():
                    line = q_stdout.get_nowait()
                    print(f"{directory}: {line}", end="")
                while not q_stderr.empty():
                    line = q_stderr.get_nowait()
                    print(f"{directory} [ERROR]: {line}", end="")
            t_stdout.join()
            t_stderr.join()

            return_code = process.wait()
            if return_code:
                raise subprocess.CalledProcessError(return_code, command)
        except subprocess.CalledProcessError as e:
            print(f"An error occurred while running '{command}' in {directory}: {e}")
        except Exception as e:
            print(f"An unexpected error occurred while running '{command}' in {directory}: {str(e)}")


if __name__ == '__main__':
    main(get_paramsdict())



import datetime
import sys
import torch.optim.lr_scheduler as _lr_scheduler
import torch
from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torchvision import datasets, transforms, models
from torch.utils.tensorboard import SummaryWriter
import os
import glob
from my_import.class_builder import ClassBuilder
import subprocess
import numpy as np

lr_scheduler = _lr_scheduler


def get_gpu():
    if torch.cuda.is_available():
        return torch.device('cuda')


class MyDataset(Dataset):
    def __init__(self, data, labels):
        self.data = data
        self.labels = labels

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx]


class MyDataLoader:

    def __init__(self, data_dir, batch=32, shuffle=True, transform=None, image_size=None):
        if image_size is None:
            image_size = 128
        if transform is None:
            transform = transforms.Compose([
                transforms.Resize((image_size, image_size)),
                transforms.ToTensor(),
                transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
            ])
        self.transform = transform
        train_dataset = datasets.ImageFolder(root=f'{data_dir}/train', transform=transform)
        val_dataset = datasets.ImageFolder(root=f'{data_dir}/val', transform=transform)
        test_dataset = datasets.ImageFolder(root=f'{data_dir}/test', transform=transform)
        self.train_loader = DataLoader(train_dataset, batch_size=batch, shuffle=shuffle)
        self.val_loader = DataLoader(val_dataset, batch_size=batch, shuffle=False)
        self.test_loader = DataLoader(test_dataset, batch_size=batch, shuffle=False)
        self.classes = train_dataset.classes
        value = next(iter(train_dataset))
        self.size = value[0].size()
        self.batch_size = batch
        print(self.size)
        self.data_dir = data_dir

    def check(self):
        for images, labels in self.train_loader:
            print(f'Image batch dimensions: {images.size()}')
            print(f'Image label dimensions: {labels.size()}')
            break

    def get_batch_index(self, index):
        test_batches = list(self.test_loader)
        return test_batches[index][0]


class Trainer:

    def __init__(self, model: nn.Module, data_loader: MyDataLoader, criterion=None, optimizer=None, device=None,
                 auto_classifer=True, log_dir=None, is_model_graph=True, auto_save=True, writer=None,
                 verbose=0, evaluation_interval=10, mem_threshold=0.9, lr=0.01, scheduler=None,
                 patience=5, save_interval=10, **kwargs):
        self.params = ClassBuilder.get_params()

        if device is None:
            device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        if isinstance(data_loader, str):
            data_loader = MyDataLoader(data_loader)
        if isinstance(model, str):
            model = torch.load(model)
        print('Using device: {device}'.format(device=device))
        if auto_classifer:
            num_ftrs = model.fc.in_features
            num_classes = len(data_loader.classes)
            model.fc = nn.Linear(num_ftrs, num_classes)
        self.device = device
        self.verbose = verbose
        self.model = model.to(device)
        self.data_loader = data_loader
        self.criterion = criterion if criterion is not None else Loss.CrossEntropyLoss()  # nn.CrossEntropyLoss()
        self.optimizer = optimizer if optimizer is not None else optim.SGD(model.parameters(), lr=lr)
        self.lr = lr
        self.save_interval = save_interval
        if callable(self.optimizer):
            self.optimizer = self.optimizer(self.model.parameters())
        if writer is None:
            writer = SummaryWriter(log_dir)
        self.log_dir = log_dir if log_dir is not None else writer.get_logdir()
        self.writer = writer
        self.auto_save = auto_save
        self.epoch = kwargs.get('epoch', 0)
        self.is_model_graph = is_model_graph
        self.evaluation_interval = evaluation_interval
        self.mem_threshold = mem_threshold
        self.scheduler = scheduler
        self.patience = patience
        self.best_loss = np.inf
        self.early_stop_counter = 0
        if is_model_graph:
            sample_data, _ = next(iter(data_loader.train_loader))
            self.add_model_graph(sample_data)

    def _early_stop(self, average_loss, epoch):
        if average_loss < self.best_loss:
            self.best_loss = average_loss
            self.early_stop_counter = 0
            self.save_model(path=f'models/best_model.pth', epoch=epoch)
        else:
            self.early_stop_counter += 1

        if self.early_stop_counter >= self.patience:
            print(f"Early stopping triggered at epoch {epoch + 1}")
            return -1
        return 0

    def add_model_graph(self, data_sample):
        self.writer.add_graph(self.model, data_sample.to(self.device))

    def train(self, num_epochs):
        if self.verbose == 2:
            self.show()
        device = self.device
        total_epochs = self.epoch + num_epochs
        outer_tqdm = tqdm(range(self.epoch, total_epochs), desc='Epochs', unit='epoch', position=0, leave=False)
        for epoch in outer_tqdm:
            self.model.train()
            running_loss = 0.0
            inner_tqdm = tqdm(self.data_loader.train_loader, desc=f'Epoch {epoch + 1}/{total_epochs}', unit='batch',
                              position=1, leave=False)
            for data, labels in inner_tqdm:
                # data, labels = self.data_processing(data, labels)
                # outputs = self._predict(data)
                # loss = self.criterion(outputs, labels)
                loss, _ = self.train_method(data, labels)

                self.backward(loss)

                loss_value = self.loss_value(loss)
                running_loss += loss_value
                inner_tqdm.set_postfix(loss=loss_value)

            mem_allocated = torch.cuda.memory_allocated(device)
            mem_reserved = torch.cuda.memory_reserved(device)
            if mem_allocated / mem_reserved > self.mem_threshold:
                tqdm.write("The memory allocation is close to the maximum, So Ending the training")
                return

            epoch_loss = running_loss / len(self.data_loader.train_loader)
            outer_tqdm.set_postfix(loss=epoch_loss)
            self.writer.add_scalar('Loss/Training', epoch_loss, epoch)
            self.writer.add_scalar('Learning Rate', self.get_lr(), epoch)
            self._auto_save(epoch)
            if self._evaluation(epoch) == -1:
                return

            self.scheduler_step()
        self.epoch += num_epochs

    def train_method(self, data, labels):
        data, labels = data.to(self.device), labels.to(self.device)
        outputs = self._predict(data)
        loss = self.criterion(outputs, labels)
        return loss, outputs

    def loss_value(self, loss):
        return loss.item()

    def backward(self, loss):
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

    @staticmethod
    def is_inverval(epoch, interval):
        return epoch % interval == 0 and epoch != 0

    def _evaluation(self, epoch):
        if Trainer.is_inverval(epoch, self.evaluation_interval):
            accuracy, avg_loss = self.evaluate()
            self.writer.add_scalar('Accuracy/Validation', accuracy, epoch)
            self.writer.add_scalar('Loss/Validation', avg_loss, epoch)
            early_stop = self._early_stop(average_loss=avg_loss, epoch=epoch)
            return early_stop
        return 0

    def scheduler_step(self):
        if self.scheduler is not None:
            self.scheduler.step()

    def _auto_save(self, epoch):
        if Trainer.is_inverval(epoch, self.save_interval):
            self.save_model(path='models/model_epoch_{:02d}.pth'.format(epoch), epoch=epoch)

    def evaluate(self, data_loader=None, phase='Validation'):
        if data_loader is None:
            data_loader = self.data_loader.val_loader
        self.model.eval()
        correct = 0
        total = 0
        running_loss = 0.0
        with torch.no_grad():
            for data, labels in tqdm(data_loader, desc='Evaluating', unit='batch'):
                loss, outputs = self.train_method(data, labels)
                running_loss += self.loss_value(loss)
                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        accuracy = 100 * correct / total
        avg_loss = running_loss / len(data_loader)
        tqdm.write(f'{phase} Accuracy: {accuracy}%')
        tqdm.write(f'{phase} Average loss: {avg_loss}')
        return accuracy, avg_loss

    def test(self):
        return self.evaluate(self.data_loader.test_loader, phase='Test')

    def _predict(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def adjust_tensor_size(self, x):
        x1, x2, x3 = self.data_loader.size[0], self.data_loader.size[1], self.data_loader.size[2]

        if x.dim() == 3 and x.size(0) == x1 and x.size(1) == x2 and x.size(2) == x3:
            x = x.unsqueeze(0)
        elif x.dim() != 4 or x.size(1) != x1 or x.size(2) != x2 or x.size(3) != x3:
            raise ValueError("输入张量的尺寸不匹配")

        return x

    def predict(self, data, transform=False, full=False):
        if transform and not isinstance(data, torch.Tensor):
            data = self.data_loader.transform(data)
        data = self.adjust_tensor_size(data)
        outputs = self._predict(data.to(self.device))
        if full:
            return F.normalize(outputs, p=1, dim=1)
        _, predicted = torch.max(outputs.data, 1)
        return predicted

    def close(self):
        self.writer.close()

    def save_model(self, path='models/model.pth', epoch=None):
        os.makedirs(os.path.dirname(path), exist_ok=True)
        if epoch is None:
            epoch = self.epoch
        torch.save({
            'epoch': epoch,
            'model_state_dict': self.model.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'loss': self.criterion,
        }, path)
        print(f'Model saved at epoch {epoch}')

    def load_model(self, path=None):
        if path is None:
            model_files = glob.glob('models/*')

            def extract_epoch(filename):
                try:
                    # 提取文件名的最后一个部分并转换为整数
                    return int(filename.split('_')[-1].split('.')[0])
                except (IndexError, ValueError):
                    # 如果提取失败，则返回一个最小值，这样这些文件不会被选中
                    return -1

            # 过滤掉无效的文件名
            valid_files = [f for f in model_files if extract_epoch(f) != -1]

            if not valid_files:
                raise ValueError("No valid model files found.")
            max_epoch_file = max(valid_files, key=extract_epoch)
            path = max_epoch_file
        if not os.path.exists(path):
            raise FileNotFoundError(f"No checkpoint found at '{path}'")
        print(f'Loading model from path {path}')
        checkpoint = torch.load(path, weights_only=False)
        self.model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        epoch = checkpoint['epoch']
        loss = checkpoint['loss']
        self.model.to(self.device)
        print(f'Model loaded from epoch {epoch}')
        return epoch, loss

    def save(self, path):
        if '.pth' in path:
            return self._save_config(path)
        os.makedirs(path, exist_ok=True)
        # self.save_model(os.path.join(path, 'model_weights.pth'))
        return self._save_config(os.path.join(path, 'trainer.pth'))

    def _save_config(self, path):

        try:
            torch.save({'model': self.model, 'epoch': self.epoch,
                        'optimizer': self.optimizer, 'criterion': self.criterion,
                        'device': self.device, 'log_dir': self.log_dir, 'verbose': self.verbose,
                        'is_model_graph': self.is_model_graph, 'auto_classifer': False,
                        'auto_save': self.auto_save, 'data_loader': self.data_loader},
                       path)
            print('Saved model configuration successfully')
            return True
        except Exception as e:
            print(f'Error saving model configuration: {str(e)}')
            return False

    @staticmethod
    def load(path):
        if not os.path.exists(path):
            raise FileNotFoundError(f"No checkpoint found at '{path}'")
        params = torch.load(os.path.join(path, 'trainer.pth'))
        return Trainer(**params)

    def show(self):
        print(self.log_dir)
        subprocess.Popen(['tensorboard', '--logdir', os.path.dirname(self.log_dir)])

        print(f"TensorBoard is running. Open http://localhost:6006/ in your browser.")

    def set_scheduler(self, scheduler, **kwargs):
        self.scheduler = scheduler(self.optimizer, **kwargs)

    def get_lr(self):
        if self.scheduler is None:
            return self.optimizer.param_groups[0]['lr']
        else:
            return self.scheduler.get_last_lr()[0]


class Models:

    @staticmethod
    def resnet18(pretrained=True):
        return models.resnet18(weights=pretrained)

    @staticmethod
    def alexnet(pretrained=True):
        return models.alexnet(weights=pretrained)

    @staticmethod
    def vgg16(pretrained=True):
        return models.vgg16(weights=pretrained)

    @staticmethod
    def densenet121(pretrained=True):
        return models.densenet121(weights=pretrained)

    @staticmethod
    def squeezenet1_0(pretrained=True):
        return models.squeezenet1_0(weights=pretrained)


class Optimizer:
    SGD = optim.SGD
    Adam = optim.Adam


class Loss:
    CrossEntropyLoss = nn.CrossEntropyLoss
    MSELoss = nn.MSELoss
    BCELoss = nn.BCELoss
    NLLLoss = nn.NLLLoss
    L1Loss = nn.L1Loss
    SmoothL1Loss = nn.SmoothL1Loss
    CTCLoss = nn.CTCLoss


class Function:
    normalize = F.normalize


class TrainAndTest:

    def __init__(self, train, test):
        self.__call__(train, test)

    def __call__(self, train, test):
        args = sys.argv[1] if len(sys.argv) > 1 else None
        if args is not None:
            if args == 'train':
                return train()
            elif args == 'test':
                return test()
        return None



from setuptools import setup, find_packages

setup(
    name="my_unique_import",
    version="0.3.16",
    packages=find_packages(),
    include_package_data=True,
    description="A custom importer package",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://pypi.org/project/my-unique-import/",
    author="Jingyuan Chen",
    author_email="jchensteve@gmail.com",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'qgit=my_import.quick_git:main',
            'EnvUpdate=my_import.update_env:main',
            'qstart=my_import.start_server:main',
            'hpy=my_import.hsr_run:main',
            'my_shutdown=my_import.auto_control:AutoControl.shutdown',
            # 'AI=my_import:AI',
        ],
    },
    python_requires='>=3.9',
    install_requires=[
        'tqdm==4.66.4'
    ],
)


import torch
import torch.nn as nn
from .AI import Trainer, MyDataLoader
from tqdm import tqdm


class _DiffusionModel(nn.Module):
    def __init__(self, base_model, image_size):
        super(_DiffusionModel, self).__init__()
        self.base_model = base_model
        self.image_size = image_size
        num_features = self.base_model.fc.in_features
        self.base_model.fc = nn.Linear(num_features, image_size // 8 * image_size // 8 * 128)

        self.deconv1 = nn.ConvTranspose2d(128, 64, kernel_size=4, stride=2, padding=1)
        self.deconv2 = nn.ConvTranspose2d(64, 32, kernel_size=4, stride=2, padding=1)
        self.deconv3 = nn.ConvTranspose2d(32, 3, kernel_size=4, stride=2, padding=1)

        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.base_model(x)
        x = x.view(x.size(0), 128, self.image_size // 8, self.image_size // 8)
        x = self.relu(self.deconv1(x))
        x = self.relu(self.deconv2(x))
        x = self.deconv3(x)
        return x


class DiffusionModel:

    @staticmethod
    def diffusion(model, image_size, device):
        return _DiffusionModel(model, image_size).to(device)


class DiffusionTrainer(Trainer):

    def __init__(self, model: nn.Module, data_loader: MyDataLoader, criterion=None, optimizer=None, device=None,
                 auto_classifer=False, log_dir=None, is_model_graph=True, auto_save=False, writer=None,
                 verbose=0, evaluation_interval=10, mem_threshold=0.9, lr=0.005, scheduler=None, save_interval=10,
                 **kwargs):
        if criterion is None:
            criterion = nn.MSELoss()

        super(DiffusionTrainer, self).__init__(model, data_loader, criterion, optimizer, device, auto_classifer,
                                               log_dir, is_model_graph, auto_save, writer, verbose, evaluation_interval,
                                               mem_threshold, lr, scheduler, save_interval, **kwargs)

    # def train(self, num_epochs):
    #     device = self.device
    #     for epoch in tqdm(range(num_epochs)):
    #         self.model.train()
    #         running_loss = 0.0
    #         for images, _ in tqdm(self.data_loader.train_loader, desc=f'Epoch {epoch + 1}/{num_epochs}',
    #                               unit='batch'):
    #             images = images.to(device)
    #             noisy_images = images + torch.randn_like(images) * 0.1
    #             self.optimizer.zero_grad()
    #             outputs = self.model(noisy_images)
    #             loss = self.criterion(outputs, images)
    #             loss.backward()
    #             self.optimizer.step()
    #             running_loss += loss.item() * images.size(0)
    #
    #         mem_allocated = torch.cuda.memory_allocated(device)
    #         mem_reserved = torch.cuda.memory_reserved(device)
    #         if mem_allocated / mem_reserved > self.mem_threshold:
    #             print("The memory allocation is close to the maximum, So Ending the training")
    #             return
    #
    #         epoch_loss = running_loss / len(self.data_loader.train_loader)
    #         print(f'Epoch [{self.epoch + 1}/{num_epochs}], Loss: {epoch_loss:.4f}')
    #         self.writer.add_scalar('Loss/Training', epoch_loss, epoch)
    #         self.writer.add_scalar('Learning Rate', self.get_lr(), epoch)
    #         self.epoch = epoch
    #         self._auto_save(epoch)
    #         if epoch % self.evaluation_interval == 0 and epoch != 0:
    #             average_loss = self.evaluate()
    #             self.writer.add_scalar('Loss/Validation', average_loss, epoch)
    #             early_stop = self._early_stop(average_loss=average_loss, epoch=epoch)
    #             if early_stop == -1:
    #                 return
    #
    #         if self.scheduler is not None:
    #             self.scheduler.step()
    def train(self, num_epochs):
        super().train(num_epochs)

    def train_method(self, data, labels):
        images = data.to(self.device)
        # tqdm.write(f'images size {images.size(0)}')
        noisy_images = images + torch.randn_like(images) * 0.1
        outputs = self.model(noisy_images)
        loss = self.criterion(outputs, images)
        return loss

    def loss_value(self, loss):
        return loss.item() * self.data_loader.batch_size

    def evaluate(self, data_loader=None, phase='Validation'):
        if data_loader is None:
            data_loader = self.data_loader.val_loader
        self.model.eval()
        running_loss = 0.0
        with torch.no_grad():
            for images, _ in tqdm(data_loader, desc='Evaluating', unit='batch'):
                images = images.to(self.device)
                noisy_images = images + torch.randn_like(images) * 0.1
                outputs = self.model(noisy_images)
                loss = self.criterion(outputs, images)
                running_loss += loss.item() * images.size(0)

        average_loss = running_loss / len(data_loader)
        print(f'{phase} Loss: {average_loss:.4f}')
        return average_loss

    def test(self):
        return self.evaluate(self.data_loader.test_loader, phase='Test')

    def _predict(self, *args, **kwargs):
        self.model.eval()
        with torch.no_grad():
            res = self.model(*args, **kwargs)
        return res

    def predict(self, data, *args, **kwargs):
        return self._predict(data.to(self.device))



import math
import random
from TTT import TicTacToe
import numpy as np


class Node:
    def __init__(self, state, parent=None, action=None):
        self.state = state
        self.parent = parent
        self.action = action
        self.children = []
        self.visits = 0
        self.value = 0

    def is_fully_expanded(self, legal_actions):
        return len(self.children) == len(legal_actions)

    def best_child(self, c_param=1.4):
        choices_weights = [
            (child.value / child.visits) + c_param * math.sqrt((2 * math.log(self.visits) / child.visits))
            for child in self.children
        ]
        return self.children[np.argmax(choices_weights)]

    def add_child(self, state, action):
        child = Node(state, self, action)
        self.children.append(child)
        return child


class MCTS:
    def __init__(self, game, n_iter=1000):
        self.game = game
        self.n_iter = n_iter


    def choice_action(self):
        legal_actions = self.game.get_legal_actions()
        action = random.choice(legal_actions)
        return action
    def search(self, initial_state):
        root = Node(initial_state)
        for _ in range(self.n_iter):
            node = self.tree_policy(root)
            reward = self.default_policy(node.state)
            self.backup(node, reward)
        return root.best_child(c_param=0)

    def tree_policy(self, node):
        while not self.is_terminal(node.state):
            legal_actions = self.game.get_legal_actions()
            if not node.is_fully_expanded(legal_actions):
                return self.expand(node, legal_actions)
            else:
                node = node.best_child()
        return node

    def expand(self, node, legal_actions):
        action = random.choice([a for a in legal_actions if a not in [child.action for child in node.children]])
        next_state, _, _ = self.game.step(action)
        return node.add_child(next_state, action)

    def default_policy(self, state):
        self.game.board = state.copy()
        current_player = self.game.current_player
        reward = 0
        while not self.is_terminal(self.game.board):
            action = self.choice_action()
            state, reward, done, _ = self.game.step(action)
            if done:
                break
        return reward if self.game.current_player != current_player else -reward

    def backup(self, node, reward):
        while node is not None:
            node.visits += 1
            node.value += reward
            node = node.parent

    def is_terminal(self, state):
        self.game.board = state
        return self.game.check_winner() != 0 or len(self.game.get_legal_actions()) == 0


# Example usage
if __name__ == '__main__':
    game = TicTacToe()
    initial_state = game.reset()
    mcts = MCTS(game, n_iter=1000)
    best_action_node = mcts.search(initial_state)
    print("Best action:", best_action_node.action)



import pygetwindow as gw
import pyautogui



class WindowSystem:

    def __init__(self):
        self.windows = self.get_windows()

    def get_windows(self):
        windows = {window.title: window for window in gw.getAllWindows()}
        self.windows = windows
        return windows

    def __repr__(self):
        return str(list(self.windows.keys()))


import numpy as np
from .Game import Game


class TicTacToe(Game):
    def __init__(self, board=None, start_player=1):
        if board is None:
            board = np.zeros((3, 3, 3), dtype=int)
        self.board = board
        self.current_player = 1
        self.start_player = start_player

    def reset(self, start_player=1):
        self.board = np.zeros((3, 3, 3), dtype=int)
        self.current_player = start_player
        self.start_player = start_player
        return self.board

    def get_legal_actions(self):
        return [(i, j) for i in range(3) for j in range(3) if self.board[0, i, j] == 0]

    def step(self, action):
        if self.board[0, action[0], action[1]] != 0:
            raise ValueError("Invalid action")
        self.board[0, action[0], action[1]] = 1
        self.board[self.current_player, action[0], action[1]] = 1
        winner = self.check_winner()
        reward = 0
        if winner == self.start_player:
            reward = 10
        elif winner == 3 - self.start_player:
            reward = -10
        elif winner == -1:
            reward = 5
        # elif winner == 0 and self.current_player == 1:
        #     reward = -1
        # elif winner == 0 and self.current_player == 2:
        #     reward = +1
        done = len(self.get_legal_actions()) == 0 or winner != 0
        self.current_player = 3 - self.current_player
        return self.board.copy(), reward, done, winner

    def check_winner(self):
        for i in range(3):
            if np.all(self.board[self.current_player, i, :] == 1):
                return self.current_player
            if np.all(self.board[self.current_player, :, i] == 1):
                return self.current_player
        if np.all(np.diag(self.board[self.current_player, :, :]) == 1):
            return self.current_player
        if np.all(np.diag(np.fliplr(self.board[self.current_player, :, :])) == 1):
            return self.current_player
        if np.all(self.board[0, :, :] == 1):
            return -1
        return 0

    def __repr__(self):
        return self.board.__repr__()

    def state(self):
        return self.board.copy()

    def __copy__(self):
        return TicTacToe(self.board.copy())

    def copy(self):
        return self.__copy__()

    def transfer_action(self, action):
        return action[0] * 3 + action[1]
